#include "GLRenderer.h"
#include "GLSetup.h"
#include <math.h>
#include "gl/gl.h"
#include "gl/glu.h"

#define VIEW_LENGTH     5.0f

GLRenderer* glrenderer;
LRESULT CALLBACK GLWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
DWORD WINAPI RendererThread(LPVOID lpParameter);

GLRenderer::GLRenderer() : parent(NULL), hwnd(NULL), created(false)
{
  ResetCamera(false);
  mousedown=false;
}

GLRenderer::~GLRenderer()
{
  glDeleteLists(cube, 1);
}

void GLRenderer::Init(HWND parent, HINSTANCE hinstance, int x, int y, int width, int height)
{
  this->parent=parent;
  this->hinstance=hinstance;
  this->x=x;
  this->y=y;
  this->width=width;
  this->height=height;

  mousedown=false;
  dispworld=dispframe=displinks=true;

  noofpoints[0]=noofpoints[1]=0;
  pointdata[0]=pointdata[1]=NULL;

  CreateThread(NULL, 0, RendererThread, this, 0, NULL);
}

DWORD WINAPI RendererThread(LPVOID lpParameter)
{
  ((GLRenderer*) lpParameter)->InitAux();
  return 0;
}

void GLRenderer::InitAux()
{
	WNDCLASSEX classname;
  classname.cbSize=sizeof(WNDCLASSEX); 
  classname.style=CS_HREDRAW|CS_VREDRAW;
  classname.lpfnWndProc=(WNDPROC) GLWndProc;
  classname.cbClsExtra=0;
  classname.cbWndExtra=0;
  classname.hInstance=hinstance;
  classname.hIcon=LoadIcon(NULL, IDI_WINLOGO);
  classname.hCursor=LoadCursor(NULL, IDC_ARROW);
  classname.hbrBackground=::GetSysColorBrush(COLOR_3DFACE);
  classname.lpszMenuName=NULL;
  classname.lpszClassName=L"GLKinectUp";
  classname.hIconSm=LoadIcon(NULL, IDI_WINLOGO);
  RegisterClassEx(&classname);

  glrenderer=this;
  hwnd=::CreateWindowEx(NULL, L"GLKinectUp", L"", WS_CHILD|WS_VISIBLE, x, y, width, height, parent, NULL, hinstance, NULL);

  InitGL(hwnd);

  glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
  glEnable(GL_CULL_FACE);
  glCullFace(GL_BACK);
  glShadeModel(GL_SMOOTH);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_NORMALIZE);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glPolygonMode(GL_FRONT, GL_FILL);

  CreateCube();
  SetupLighting();

  MSG msg;
  do
  {
    if (GetMessage(&msg, NULL, 0, 0))
    {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
  } while (msg.message!=WM_QUIT);

}

LRESULT CALLBACK GLWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{

  switch (message) 
  {
  case WM_CREATE:
    glrenderer->created=true;
    return 0;
    break;
  case WM_SIZE:
    glrenderer->OnSize((DWORD) wParam, (UINT) (lParam&0xFFFF), (UINT) (lParam>>16));
    break;
  case WM_PAINT:
    {
      PAINTSTRUCT paint;
      BeginPaint(hWnd, &paint);
      glrenderer->Draw();
      EndPaint(hWnd, &paint);
    }
    break;
  case WM_TIMER:
    break;
  case WM_DESTROY:
    PostQuitMessage(0);
    break;
  case WM_CLOSE:
    DestroyWindow(hWnd);
    //PostQuitMessage(0);
    break;
  case WM_LBUTTONDOWN:
    glrenderer->OnLButtonDown(wParam, lParam&0xFFFF, lParam>>16);
    break;
  case WM_MOUSEMOVE:
    glrenderer->OnMouseMove(wParam, lParam&0xFFFF, lParam>>16);
    break;
  case WM_LBUTTONUP:
    glrenderer->OnLButtonUp(wParam, lParam&0xFFFF, lParam>>16);
    break;
  case WM_MOUSEWHEEL:
    glrenderer->OnMouseWheel(wParam, lParam&0xFFFF, lParam>>16);
    return 0;
    break;
  case WM_KEYDOWN:
    //mainwnd->OnKeyDown(wParam, lParam&0xFFFF, lParam>16);
    break;
  case WM_COMMAND:
    {
/*      WORD wmID=(WORD) (wParam&0xFFFF);
      WORD wmevent=(WORD) (wParam>>16);
      if (wmevent==CBN_SELCHANGE)
        mainwnd->OnComboChange();*/
    }
    break;
/*  case WM_KEYDOWN:
    mainwnd->OnKeyDown(wParam, lParam&0xFFFF, lParam>>16);
    break;
  case WM_KEYUP:
    mainwnd->OnKeyUp(wParam, lParam&0xFFFF, lParam>>16);
    break;*/
  case GLREDRAW:
    glrenderer->Draw();
    break;
  default:
    return DefWindowProc(hWnd, message, wParam, lParam);
  }
  return DefWindowProc(hWnd, message, wParam, lParam);
}

void GLRenderer::SetPointData(POINTDATA* pointdata, const int noofpoints, const bool redraw)
{
  this->pointdata[0]=pointdata;
  this->noofpoints[0]=noofpoints;
  if (redraw)
    SendMessage(hwnd, GLREDRAW, 0, 0);
}

void GLRenderer::SetMatchData(POINTDATA* pointdata, const int noofpoints, const bool redraw)
{
  this->pointdata[1]=pointdata;
  this->noofpoints[1]=noofpoints;
  if (redraw)
    SendMessage(hwnd, GLREDRAW, 0, 0);
}

void GLRenderer::ResetCamera(bool redraw)
{
  eye[0]=0.0f;
  eye[1]=0.0f;
  eye[2]=VIEW_LENGTH;
 
  rotation[0]=rotation[2]=0.0f;
  rotation[1]=(float) M_PI;

  CalcLookAt(true);

  if (redraw)
    SendMessage(hwnd, GLREDRAW, 0, 0);
}

void GLRenderer::CalcLookAt(bool fromeye)
{
  float mat[16];
  CalcRotationMatrix(mat);

  if (fromeye)
  {
    centre[0]=eye[0]+mat[8]*VIEW_LENGTH;
    centre[1]=eye[1]+mat[9]*VIEW_LENGTH;
    centre[2]=eye[2]+mat[10]*VIEW_LENGTH;
  }
  else
  {
    eye[0]=centre[0]-mat[8]*VIEW_LENGTH;
    eye[1]=centre[1]-mat[9]*VIEW_LENGTH;
    eye[2]=centre[2]-mat[10]*VIEW_LENGTH;
  }

  up[0]=mat[4];
  up[1]=mat[5];
  up[2]=mat[6];
}

void GLRenderer::CalcRotationMatrix(float mat[16])
{
  float cx=(float) cos(rotation[0]);
  float sx=(float) sin(rotation[0]);

  float cy=(float) cos(rotation[1]);
  float sy=(float) sin(rotation[1]);

  float cz=(float) cos(rotation[2]);
  float sz=(float) sin(rotation[2]);

  // 0-3 = first column, etc
  mat[0]=cy*cz;
  mat[1]=sx*sy*cz+cx*sz;
  mat[2]=-cx*sy*cz+sx*sz;
  mat[3]=0.0f;

  mat[4]=-cy*sz;
  mat[5]=-sx*sy*sz+cx*cz;
  mat[6]=cx*sy*sz+sx*cz;
  mat[7]=0.0f;

  mat[8]=sy;
  mat[9]=-sx*cy;
  mat[10]=cx*cy;
  mat[11]=0.0f;

  mat[12]=0.0f;
  mat[13]=0.0f;
  mat[14]=0.0f;
  mat[15]=1.0f;
}


void GLRenderer::Draw()
{
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  glLoadIdentity();
  gluLookAt(eye[0], eye[1], eye[2], centre[0], centre[1], centre[2], up[0], up[1], up[2]);

  glPushMatrix();
    glDisable(GL_LIGHTING);
    glLineWidth(2.0);
    glBegin(GL_LINES);
      glColor3f(1.0f, 0.0f, 0.0f);
      glVertex3f(0.0f, 0.0f, 0.0f);
      glVertex3f(0.2f, 0.0f, 0.0f);

      glColor3f(0.0f, 1.0f, 0.0f);
      glVertex3f(0.0f, 0.0f, 0.0f);
      glVertex3f(0.0f, 0.2f, 0.0f);

      glColor3f(0.0f, 0.0f, 1.0f);
      glVertex3f(0.0f, 0.0f, 0.0f);
      glVertex3f(0.0f, 0.0f, 0.2f);
    glEnd();
    glEnable(GL_LIGHTING);
  glPopMatrix();

  glPushMatrix();
    DrawPointCloud();
  glPopMatrix();


  glFlush();
  SwapBuffers(wglGetCurrentDC());

}

void GLRenderer::DrawPointCloud()
{
  const float PAIRED_COLOUR_WORLD[3]={0.0f, 1.0f, 0.0f};
  const float NON_PAIRED_COLOUR_WORLD[3]={0.0f, 0.0f, 1.0f};

  const float PAIRED_COLOUR_FRAME[3]={0.0f, 1.0f, 1.0f};
  const float NON_PAIRED_COLOUR_FRAME[3]={1.0f, 0.0f, 0.0f};
  glDisable(GL_LIGHTING);
  glPointSize(4.0f);
  glBegin(GL_POINTS);
  if (dispworld)
  {
    for (int i=0; i<noofpoints[0]; i++)
    {
      if (pointdata[0][i].pairid>0)
        glColor3fv(PAIRED_COLOUR_WORLD);
      else
        glColor3fv(NON_PAIRED_COLOUR_WORLD);
      glVertex3fv(pointdata[0][i].xyz);
    }
  }

  if (dispframe)
  {
    for (int i=0; i<noofpoints[1]; i++)
    {
      if (pointdata[1][i].pairid>0)
        glColor3fv(PAIRED_COLOUR_FRAME);
      else
        glColor3fv(NON_PAIRED_COLOUR_FRAME);
      glVertex3fv(pointdata[1][i].xyz);
    }
  }

  glEnd();


  const float SCALE=0.02f;
  glBegin(GL_LINES);

//  glColor3f(0.7f, 0.7f, 0.7f);
  glLineWidth(1.0f);
  for (int i=0; i<2; i++)
  {
    if ((i==0 && dispworld) || (i==1 && dispframe))
    {
      for (int j=0; j<noofpoints[i]; j++)
      {
        if (pointdata[i][j].pairid>0)
          glColor3fv(i==0?PAIRED_COLOUR_WORLD:PAIRED_COLOUR_FRAME);
        else
          glColor3fv(i==0?NON_PAIRED_COLOUR_WORLD:NON_PAIRED_COLOUR_FRAME);

        glVertex3fv(pointdata[i][j].xyz);
        glVertex3f(pointdata[i][j].xyz[0]+pointdata[i][j].nxyz[0]*SCALE, pointdata[i][j].xyz[1]+pointdata[i][j].nxyz[1]*SCALE, pointdata[i][j].xyz[2]+pointdata[i][j].nxyz[2]*SCALE);
      }
    }
  }

  if (displinks)
  {
    glColor3f(1.0f, 1.0f, 1.0f);
    glLineWidth(2.0f);

    int to, from;
    if (toworld)
    {
      from=1;
      to=0;
    }
    else
    {
      from=0;
      to=1;
    }
    
    for (int i=0; i<noofpoints[from]; i++)
    {
      if (pointdata[from][i].pairid>=0)
      {
        glVertex3fv(pointdata[from][i].xyz);
        glVertex3fv(pointdata[to][pointdata[from][i].pairid].xyz);
      }
    }
  }

  glEnd();

  glEnable(GL_LIGHTING);

/*    glDisable(GL_LIGHTING);
      glPointSize(4.0f);
      glBegin(GL_POINTS);
        glColor3f(1.0f, 1.0f, 1.0f);
        glVertex3f(1.0f, 1.0f, 0.0f);

      glEnd();
    glEnable(GL_LIGHTING);*/
  glPopMatrix();
}
   
void GLRenderer::OnSize(UINT nType, int cx, int cy)
{
  const float NEAR_CLIP=0.01f;
  const float FAR_CLIP=100.0f;
  const float FOVY=(60.0f*(float)M_PI/180.0f);

  float fAspect, top, bottom, left, right;

  glViewport(0, 0, width, height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  if (height!=0)
    fAspect=(float) width/height;
  else
    fAspect=1.0f;

  top=(float) tan(FOVY*0.5)*NEAR_CLIP;
  bottom=-top;
  left=fAspect*bottom;
  right=fAspect*top;

//  float scale=2.0f/(right-left);
  
  glFrustum(left, right, bottom, top, NEAR_CLIP, FAR_CLIP);
  glMatrixMode(GL_MODELVIEW);

  SendMessage(hwnd, GLREDRAW, 0, 0);
}

void GLRenderer::SetWindowPos(int x, int y, int width, int height) 
{
  this->x=x;
  this->y=y;
  this->width=width;
  this->height=height;

  ::SetWindowPos(hwnd, NULL, x, y, width, height, SWP_NOZORDER);
}

void GLRenderer::GetWindowPos(int& x, int& y, int& width, int&height)
{
  x=this->x;
  y=this->y;
  width=this->width;
  height=this->height;
}
  
void GLRenderer::SetupLighting()
{
  float POSITION[]={0.0f, 5.f, 5.0f, 1.0f};
  float AMBIENT[]={0.1f, 0.1f, 0.1f, 1.0f};
  float DIFFUSE[]={0.8f, 0.8f, 0.8f, 1.0f};
  float SPECULAR[]={0.1f, 0.1f, 0.1f, 1.0f};

  glLightfv(GL_LIGHT0, GL_POSITION, POSITION);
  glLightfv(GL_LIGHT0, GL_AMBIENT, AMBIENT);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, DIFFUSE);
  glLightfv(GL_LIGHT0, GL_SPECULAR, SPECULAR);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
}

void GLRenderer::OnLButtonDown(UINT nType, int x, int y)
{
  mousedown=true;
  lastpoint.x=x;
  lastpoint.y=y;
}

void GLRenderer::OnMouseMove(UINT nType, int x, int y)
{
  float SCALE=0.01f;
  if (nType&MK_SHIFT)
    SCALE*=0.1f;

  if (nType&MK_LBUTTON)
  {
    if (mousedown)
    {
      int dx=x-lastpoint.x;
      int dy=y-lastpoint.y;
      if (nType&MK_CONTROL)
      {
        rotation[0]-=dy*SCALE;
        rotation[1]-=dx*SCALE;
        CalcLookAt(false);
      }
      else
      {
        float mat[16];
        CalcRotationMatrix(mat);

        float tx=(dx*mat[0]+dy*mat[4])*SCALE;
        float ty=(dx*mat[1]+dy*mat[5])*SCALE;
        float tz=(dx*mat[2]+dy*mat[6])*SCALE;

        eye[0]+=tx;
        eye[1]+=ty;
        eye[2]+=tz;

        centre[0]+=tx;
        centre[1]+=ty;
        centre[2]+=tz;
      }

      lastpoint.x=x;
      lastpoint.y=y;
      Draw();
    }
  }
  else if (mousedown)
    mousedown=false;
}

void GLRenderer::OnMouseWheel(WPARAM wParam, int x, int y)
{
  float SCALE=0.01f;
  short amt=(short) (wParam>>16);

  if (wParam&MK_SHIFT)
    SCALE*=0.01f;

  float mat[16];
  CalcRotationMatrix(mat);

  float tx=(amt*mat[8])*SCALE;
  float ty=(amt*mat[9])*SCALE;
  float tz=(amt*mat[10])*SCALE;

  eye[0]+=tx;
  eye[1]+=ty;
  eye[2]+=tz;

  centre[0]+=tx;
  centre[1]+=ty;
  centre[2]+=tz;

  SendMessage(hwnd, GLREDRAW, 0, 0);
}

void GLRenderer::OnLButtonUp(UINT nType, int x, int y)
{
  mousedown=false;
}

void GLRenderer::CreateCube()
{
  cube=glGenLists(1);

  glNewList(cube, GL_COMPILE);
    glScalef(0.02f, 0.02f, 0.02f);
//    glScalef(0.08f, 0.08f, 0.08f);
    glBegin(GL_QUADS);
      glNormal3f(0.0f, 0.0f, 1.0f);
      glVertex3f(-0.5f, 0.5f, 0.5f);
      glVertex3f(-0.5f, -0.5f, 0.5f);
      glVertex3f(0.5f, -0.5f, 0.5f);
      glVertex3f(0.5f, 0.5f, 0.5f);

      glNormal3f(0.0f, 0.0f, -1.0f);
      glVertex3f(0.5f, 0.5f, -0.5f);
      glVertex3f(0.5f, -0.5f, -0.5f);
      glVertex3f(-0.5f, -0.5f, -0.5f);
      glVertex3f(-0.5f, 0.5f, -0.5f);

      glNormal3f(1.0f, 0.0f, 0.0f);
      glVertex3f(0.5f, 0.5f, 0.5f);
      glVertex3f(0.5f, -0.5f, 0.5f);
      glVertex3f(0.5f, -0.5f, -0.5f);
      glVertex3f(0.5f, 0.5f, -0.5f);

      glNormal3f(-1.0f, 0.0f, 0.0f);
      glVertex3f(-0.5f, 0.5f, -0.5f);
      glVertex3f(-0.5f, -0.5f, -0.5f);
      glVertex3f(-0.5f, -0.5f, 0.5f);
      glVertex3f(-0.5f, 0.5f, 0.5f);

      glNormal3f(0.0f, 1.0f, 0.0f);
      glVertex3f(0.5f, 0.5f, 0.5f);
      glVertex3f(0.5f, 0.5f, -0.5f);
      glVertex3f(-0.5f, 0.5f, -0.5f);
      glVertex3f(-0.5f, 0.5f, 0.5f);

      glNormal3f(0.0f, -1.0f, 0.0f);
      glVertex3f(-0.5f, -0.5f, 0.5f);
      glVertex3f(-0.5f, -0.5f, -0.5f);
      glVertex3f(0.5f, -0.5f, -0.5f);
      glVertex3f(0.5f, -0.5f, 0.5f);
    glEnd();
  glEndList();
}
