#pragma once
#include <Windows.h>

#define GLREDRAW        WM_USER+1

struct POINTDATA
{
  int pairid, index;
  float xyz[3], nxyz[3];
};

class GLRenderer
{
public:
  GLRenderer();
  ~GLRenderer();

  HWND GetHwnd();

  void Init(HWND parent, HINSTANCE hinstance, int x, int y, int width, int height);
  void OnSize(UINT nType, int cx, int cy);
  void SetWindowPos(int x, int y, int width, int height);
  void GetWindowPos(int& x, int& y, int& width, int&height);
  void Draw();
  bool HitTest(int x, int y);
  void ResetCamera(bool redraw);
  bool Created();

  void OnLButtonDown(UINT nType, int x, int y);
  void OnMouseMove(UINT nType, int x, int y);
  void OnLButtonUp(UINT nType, int x, int y);
  void OnMouseWheel(WPARAM wParam, int x, int y);

  void ToggleWorldData();
  void ToggleFrameData();
  void ToggleLinks();

  void SetToWorld(const bool toworld);
  void SetPointData(POINTDATA* pointdata, const int noofpoints, const bool redraw);
  void SetMatchData(POINTDATA* pointdata, const int noofpoints, const bool redraw);

  HWND parent, hwnd;

private:
  HINSTANCE hinstance;
  int x, y, width, height;
  bool created;
  
  float rotation[3];
  float eye[3], centre[3], up[3];
      
  bool mousedown;
  POINT lastpoint;

  bool dispworld, dispframe, displinks;

  unsigned int cube;
  
  bool toworld;
  int noofpoints[2];
  POINTDATA* pointdata[2];

  void SetupLighting();
  void DrawPointCloud();
  void DrawCameraInfo();
  void CalcRotationMatrix(float mat[16]);
  void CalcLookAt(bool fromeye);

  void InitAux();
  void CreateCube();

  friend LRESULT CALLBACK GLWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
  friend DWORD WINAPI RendererThread(LPVOID lpParameter);
};

inline bool GLRenderer::HitTest(int x, int y)
{
  return (x>=this->x && x<=(this->x+width) && y>=this->y && y<=(this->y+height));
}

inline bool GLRenderer::Created()
{
  return created;
}

inline void GLRenderer::ToggleWorldData()
{
  dispworld^=true;
  SendMessage(hwnd, GLREDRAW, 0, 0);
}

inline void GLRenderer::ToggleFrameData()
{
  dispframe^=true;
  SendMessage(hwnd, GLREDRAW, 0, 0);
}

inline void GLRenderer::ToggleLinks()
{
  displinks^=true;
  SendMessage(hwnd, GLREDRAW, 0, 0);
}

inline void GLRenderer::SetToWorld(const bool toworld)
{
  this->toworld=toworld;
}

inline HWND GLRenderer::GetHwnd()
{
  return hwnd;
}
