#include "GLSetup.h"
#include <assert.h>
#include <Math.h>

BOOL bSetupPixelFormat(HDC hdc);

HDC InitGL(HWND parent)
{
  RECT rect;
	GLfloat	fAspect;
  HDC hdcNew=::GetDC(parent);

  assert(hdcNew);

  if (!bSetupPixelFormat(hdcNew))
    return 0;

  HGLRC hrc=wglCreateContext(hdcNew);
  wglMakeCurrent(hdcNew, hrc);

  glClearDepth(1.0f);
  glEnable(GL_DEPTH_TEST);
  ::GetClientRect(parent, &rect);
  fAspect=(GLfloat)rect.right/rect.bottom;

  glPolygonOffset(-2.0, -1.0);
  
  glViewport(0, 0, rect.right, rect.bottom);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  
  float top=(float) tan(0.4f)*0.1f;
  float bottom=-top;
  float left=fAspect*bottom;
  float right=fAspect*top;
  
  glFrustum(left, right, bottom, top, 0.1f, 100.0f);
  glMatrixMode(GL_MODELVIEW);
  return hdcNew;
}

BOOL bSetupPixelFormat(HDC hdc)
{
  int pixelformat;
  PIXELFORMATDESCRIPTOR pfd = 
	{
    sizeof(PIXELFORMATDESCRIPTOR),  // size of this pfd
    1,                              // version number
    PFD_DRAW_TO_WINDOW |            // support window
    PFD_SUPPORT_OPENGL |            // support OpenGL
    PFD_DOUBLEBUFFER,               // double buffered
    PFD_TYPE_RGBA,                  // RGBA type
    24,                             // 24-bit color depth
    0, 0, 0, 0, 0, 0,               // color bits ignored
    0,                              // no alpha buffer
    0,                              // shift bit ignored
    0,                              // no accumulation buffer
    0, 0, 0, 0,                     // accum bits ignored
    32,                             // 32-bit z-buffer
    32,                             // 32-bit stencil buffer  - !no stencil buffer
    0,                              // no auxiliary buffer
    PFD_MAIN_PLANE,                 // main layer
    0,                              // reserved
    0, 0, 0                         // layer masks ignored
  };

  if (!(pixelformat=ChoosePixelFormat(hdc, &pfd)))
  {
    MessageBox(NULL, L"ChoosePixelFormat failed", L"GLSetup", MB_OK);
    return FALSE;
  }

  PIXELFORMATDESCRIPTOR descriptor;
  memset(&descriptor, 0, sizeof(PIXELFORMATDESCRIPTOR));
  descriptor.nSize=sizeof(PIXELFORMATDESCRIPTOR);
  DescribePixelFormat(hdc, pixelformat, descriptor.nSize, &descriptor);

  if (!(descriptor.dwFlags&PFD_GENERIC_ACCELERATED) && (descriptor.dwFlags&PFD_GENERIC_FORMAT))
    MessageBox(NULL, L"No hardware acceleration", L"GLSetup", MB_OK);

  if (!SetPixelFormat(hdc, pixelformat, &pfd))
  {
    MessageBox(NULL, L"SetPixelFormat failed", L"GLSetup", MB_OK);
    return FALSE;
  }
  return TRUE;
}
/*
void CreateTextureMap(const wchar_t* filename, UINT& texturemap, UINT clampmode, float alpha, bool flip)
{
  BITMAP bm;
  BYTE* imgdata=NULL;
  DWORD* temp, *data=NULL;
  int width, height;

  const wchar_t* ext=wcsrchr(filename, '.');
  if (ext)
  {
    if (!_wcsicmp(ext, L".bmp"))
    {
      HBITMAP hbmp=(HBITMAP) LoadImage(NULL, filename, IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION|LR_LOADFROMFILE);
      ::GetObject(hbmp, sizeof(BITMAP), &bm);
      imgdata=(BYTE*) bm.bmBits;
      temp=data=(DWORD*) malloc(bm.bmWidth*bm.bmHeight*4);

      DWORD trans=(DWORD) (alpha*255)<<24;
      
      for (int j=0; j<bm.bmHeight; j++)
        for (int i=0; i<bm.bmWidth; i++)
        {
          *(temp++)=(DWORD) (trans|(imgdata[0]<<16)|(imgdata[1]<<8)|imgdata[2]);
          imgdata+=3;
        }

      ::DeleteObject(hbmp);
      width=bm.bmWidth;
      height=bm.bmHeight;
    }
    else if (!_wcsicmp(ext, L".jpg"))
    {
      temp=data=(DWORD*) ReadJPEG(filename, width, height, flip);
      if (data)
      {
        DWORD trans=(DWORD) (alpha*255)<<24;
        for (int i=width*height; i>0; i--)
          *(temp++)|=trans;
      }
    }
  }
  CreateTextureMap(data, width, height, texturemap, clampmode);
  free(data);
}

void CreateTextureMap(DWORD* data, int width, int height, UINT& texturemap, UINT clampmode)
{
  if (data)
  {
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glGenTextures(1, &texturemap);
    glBindTexture(GL_TEXTURE_2D, texturemap);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, clampmode);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, clampmode);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_BGRA_EXT, GL_UNSIGNED_BYTE, data);
  }
  else
    texturemap=-1;
}*/