#include "MainWnd.h"
#include <assert.h>

MainWnd* mainwnd;
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);

#define CAPTURE_FOLDER  L"PairView_Capture\\PairView"

MainWnd::MainWnd(HINSTANCE hInstance, LPSTR lpCmdLine) : hinstance(hInstance), offsets(NULL), maxwldpts(0), worldpoints(NULL), maxmatchpts(0), matchpoints(NULL)
{
	WNDCLASSEX classname;
  classname.cbSize=sizeof(WNDCLASSEX); 
  classname.style=CS_HREDRAW|CS_VREDRAW;
  classname.lpfnWndProc=(WNDPROC) WndProc;
  classname.cbClsExtra=0;
  classname.cbWndExtra=0;
  classname.hInstance=hInstance;
  classname.hIcon=LoadIcon(NULL, IDI_WINLOGO);
  classname.hCursor=LoadCursor(NULL, IDC_ARROW);
  classname.hbrBackground=::GetSysColorBrush(COLOR_3DFACE);
  classname.lpszMenuName=NULL;
  classname.lpszClassName=L"PairViewer";
  classname.hIconSm=LoadIcon(NULL, IDI_WINLOGO);
  RegisterClassEx(&classname);

  int width=800;
  int height=640;
  int offx=(::GetSystemMetrics(SM_CXSCREEN)-width)/2;
  int offy=(::GetSystemMetrics(SM_CYSCREEN)-height)/2;

  filename[0]='\0';
  if (lpCmdLine && lpCmdLine[0])
  {
    if (lpCmdLine[0]=='\"')
      strcpy_s(filename, MAX_PATH, lpCmdLine+1);
    else
      strcpy_s(filename, MAX_PATH, lpCmdLine);

    int len=(int) strlen(filename);
    if (filename[len-1]=='\"')
      filename[len-1]='\0';
  }
  if (!filename[0])
    strcpy_s(filename, MAX_PATH, "MatchInfo.dat");
//    strcpy_s(filename, MAX_PATH, "C:\\Work\\University\\HRI\\Kinecting\\KinectUp\\MatchInfo.dat");

  ::mainwnd=this;
  hwnd=CreateWindowEx(NULL, L"PairViewer", L"Kinecting up the Past - PairViewer", WS_OVERLAPPEDWINDOW|WS_VISIBLE, offx, offy, width, height, NULL, NULL, hInstance, NULL);

  MSG msg;
  do
  {
    if (GetMessage(&msg, NULL, 0, 0))
    {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
  } while (msg.message!=WM_QUIT);

  renderer.SetPointData(NULL, 0, false);

  if (file)
    fclose(file);
  free(offsets);

  free(worldpoints);
  free(matchpoints);
}

MainWnd::~MainWnd()
{

}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  switch (message) 
  {
  case WM_CREATE:
    mainwnd->Init(hWnd);
    return 0;
    break;
  case WM_SIZE:
    mainwnd->OnSize((DWORD) wParam, (UINT) (lParam&0xFFFF), (UINT) (lParam>>16));
    break;
  case WM_PAINT:
    {
      PAINTSTRUCT paint;
      BeginPaint(hWnd, &paint);
      mainwnd->OnDraw();
      EndPaint(hWnd, &paint);
    }
    break;
  case WM_TIMER:
    mainwnd->OnTimer(wParam);
    break;
  case WM_DESTROY:
    PostQuitMessage(0);
    break;
  case WM_CLOSE:
    DestroyWindow(hWnd);
    PostQuitMessage(0);
    break;
  case WM_LBUTTONDOWN:
    mainwnd->OnLButtonDown(wParam, lParam&0xFFFF, lParam>>16);
    break;
  case WM_MOUSEMOVE:
    mainwnd->OnMouseMove(wParam, lParam&0xFFFF, lParam>>16);
    break;
  case WM_LBUTTONUP:
    mainwnd->OnLButtonUp(wParam, lParam&0xFFFF, lParam>>16);
    break;
  case WM_MOUSEWHEEL:
    mainwnd->OnMouseWheel(wParam, lParam&0xFFFF, lParam>>16);
    return 0;
    break;
  case WM_KEYDOWN:
    mainwnd->OnKeyDown(wParam, lParam&0xFFFF, lParam>16);
    break;
  case WM_COMMAND:
    {
/*      WORD wmID=(WORD) (wParam&0xFFFF);
      WORD wmevent=(WORD) (wParam>>16);
      if (wmevent==CBN_SELCHANGE)
        mainwnd->OnComboChange();*/
    }
    break;
/*  case WM_KEYDOWN:
    mainwnd->OnKeyDown(wParam, lParam&0xFFFF, lParam>>16);
    break;
  case WM_KEYUP:
    mainwnd->OnKeyUp(wParam, lParam&0xFFFF, lParam>>16);
    break;*/
  case WM_HSCROLL:
    mainwnd->OnHScroll(wParam, lParam);
    return 0;
    break;
  case GLREDRAW:
    mainwnd->OnGLRedraw(lParam);
    break;
  default:
    return DefWindowProc(hWnd, message, wParam, lParam);
  }
  return DefWindowProc(hWnd, message, wParam, lParam);
}

void MainWnd::Init(HWND hwnd)
{
  this->hwnd=hwnd;

  HDC hdc=::GetDC(NULL);
  LOGFONT logfont;
  logfont.lfHeight=-(logfont.lfHeight=-MulDiv(10, GetDeviceCaps(hdc, LOGPIXELSY), 72));
  logfont.lfWidth=0;
  logfont.lfEscapement=0;
  logfont.lfOrientation=0;
  logfont.lfWeight=FW_NORMAL;
  logfont.lfItalic=FALSE;
  logfont.lfUnderline=FALSE;
  logfont.lfStrikeOut=FALSE;
  logfont.lfCharSet=DEFAULT_CHARSET;
  logfont.lfOutPrecision=OUT_DEFAULT_PRECIS;
  logfont.lfClipPrecision=CLIP_DEFAULT_PRECIS;
  logfont.lfQuality=DEFAULT_QUALITY;
  logfont.lfPitchAndFamily=DEFAULT_PITCH;
  wcscpy_s(logfont.lfFaceName, 32, L"Tahoma");
  HFONT hfont=CreateFontIndirect(&logfont);

  info=CreateWindowEx(WS_EX_CLIENTEDGE, L"EDIT", L"", WS_CHILD|WS_VISIBLE|WS_VSCROLL|ES_AUTOVSCROLL|ES_MULTILINE|ES_READONLY, 250, 350, 300, 100, hwnd, NULL, hinstance, NULL);
  SendMessage(info, WM_SETFONT, (WPARAM) hfont, FALSE);
  SendMessage(info, EM_SETLIMITTEXT, 0, 0);

  scroll=CreateWindow(L"SCROLLBAR", L"", WS_CHILD|WS_VISIBLE|WS_VSCROLL|SBS_HORZ, 250, 350, 300, 100, hwnd, NULL, hinstance, NULL);

  ::ReleaseDC(NULL, hdc);

  renderer.Init(hwnd, hinstance, 0, 410, 240, 180);

  while (!renderer.Created())
    Sleep(50);

  file=NULL;
  fopen_s(&file, filename, "rb");
  if (file)
  {
    noofframes=0;
    int pts;
    char toworld;
    while (fread(&pts, sizeof(int), 1, file)==1)  // tracked frame number
    {
      if (noofframes==870)
        noofframes=noofframes;
      offsets=(unsigned __int64*) realloc(offsets, sizeof(unsigned __int64)*(noofframes+1));
      offsets[noofframes++]=_ftelli64(file);

      fread(&toworld, 1, 1, file);

      fread(&pts, sizeof(int), 1, file);
//      if (pts<0 || pts>100000)
//        pts=pts;
      int skip=6*sizeof(float);
      if (toworld)
        skip+=sizeof(int);
      _fseeki64(file, pts*skip, SEEK_CUR);

      fread(&pts, sizeof(int), 1, file);
      skip=6*sizeof(float)+sizeof(int);
      if (toworld)
        skip+=sizeof(int);
      _fseeki64(file, pts*skip, SEEK_CUR);
    }
    _fseeki64(file, 0, SEEK_CUR);

    wchar_t msg[MAX_PATH];
    _itow_s(noofframes, msg, MAX_PATH, 10);
    wcscat_s(msg, MAX_PATH, L" frames\r\n");
    AppendInfoText(msg);

    SetScrollRange(scroll, SB_CTL, 1, noofframes, FALSE);
    SetScrollPos(scroll, SB_CTL, 1, TRUE);
    LoadFrame(1);
  }
  else
    MessageBox(hwnd, L"Cannot open file", L"PairViewer", MB_ICONERROR);
}

void MainWnd::OnSize(UINT nType, int cx, int cy)
{
  renderer.SetWindowPos(0, 0, cx, cy-150);
  SetWindowPos(info, NULL, 0, cy-150, cx, 135, SWP_NOZORDER);
  SetWindowPos(scroll, NULL, 0, cy-15, cx, 15, SWP_NOZORDER);
}

void MainWnd::OnTimer(UINT nIDEvent)
{

}

void MainWnd::AppendInfoText(const wchar_t* text)
{
  int textlen=GetWindowTextLength(info);
  SendMessage(info, EM_SETSEL, textlen, textlen);
  SendMessage(info, EM_REPLACESEL, FALSE, (LPARAM) text);

  int textlen2=GetWindowTextLength(info);
  if (textlen==textlen2)
  {
    SetWindowText(info, text);
    textlen2=wcslen(text);
  }

  SendMessage(info, EM_SETSEL, textlen2, textlen2);
}

void MainWnd::OnLButtonDown(UINT nFlags, int x, int y)
{
}

void MainWnd::OnMouseMove(UINT nFlags, int x, int y)
{
}

void MainWnd::OnLButtonUp(UINT nFlags, int x, int y)
{
}

void MainWnd::OnMouseWheel(WPARAM wParam, int x, int y)
{
  RECT rect;
  ::GetWindowRect(hwnd, &rect);
  x-=rect.left+GetSystemMetrics(SM_CXSIZEFRAME);
  y-=rect.top+GetSystemMetrics(SM_CYSIZEFRAME)+GetSystemMetrics(SM_CYCAPTION);
  if (renderer.HitTest(x, y))
    renderer.OnMouseWheel(wParam, x, y);
}

void MainWnd::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
{
  if (nChar=='R')
    renderer.ResetCamera(true);
  else if (nChar=='W')
    renderer.ToggleWorldData();
  else if (nChar=='F')
    renderer.ToggleFrameData();
  else if (nChar=='L')
    renderer.ToggleLinks();
  else if (nChar=='C')
    CaptureVideo();
  else if (nChar==VK_LEFT)
    OnHScroll(SB_LINELEFT, 0);
  else if (nChar==VK_RIGHT)
    OnHScroll(SB_LINERIGHT, 0);
}

void MainWnd::OnDraw()
{
  OnGLRedraw(3);
}

void MainWnd::OnGLRedraw(LPARAM lParam)
{
  SendMessage(renderer.hwnd, GLREDRAW, 0, 0);
}

void MainWnd::OnHScroll(WPARAM wParam, LPARAM lParam)
{
  WORD req=wParam&0xFFFF;
  WORD val=(wParam>>16)&0xFFFF;

//  int min, max;
//  GetScrollRange(scroll, SB_HORZ, &min, &max);

  int pos, curpos=GetScrollPos(scroll, SB_CTL);
  if (req==SB_THUMBPOSITION || req==SB_THUMBTRACK)
    pos=val;
  else
  {
    pos=curpos;
    int page=noofframes/20;
    if (page==0)
      page=1;

    if (req==SB_LEFT)
      pos=0;
    else if (req==SB_RIGHT)
      pos=noofframes;
    else if (req==SB_LINELEFT)
      --pos;
    else if (req==SB_LINERIGHT)
      ++pos;
    else if (req==SB_PAGELEFT)
      pos-=page;
    else if (req==SB_PAGERIGHT)
      pos+=page;

    if (pos<1)
      pos=1;
    else if (pos>noofframes)
      pos=noofframes;
  }

  if (curpos!=pos)
  {
    SetScrollPos(scroll, SB_CTL, pos, true);
    LoadFrame(pos);
  }
}

void MainWnd::LoadFrame(int frame)
{
  if (file)
  {
    if (frame<1)
      frame=1;
    else if (frame>noofframes)
      frame=noofframes;

    renderer.SetPointData(NULL, 0, false);
    renderer.SetMatchData(NULL, 0, false);

    int i, noofwldpts, noofmatchpts;
    char tmp;

    int matches=0;

    _fseeki64(file, offsets[frame-1], SEEK_SET);
    fread(&tmp, 1, 1, file);
    toworld=(tmp==0);

    fread(&noofwldpts, sizeof(int), 1, file);
    if (noofwldpts>maxwldpts)
    {
      worldpoints=(POINTDATA*) realloc(worldpoints, noofwldpts*sizeof(POINTDATA));
      maxwldpts=noofwldpts;
    }
    for (i=0; i<noofwldpts; i++)
    {
      fread(worldpoints[i].xyz, sizeof(float), 3, file);
      fread(worldpoints[i].nxyz, sizeof(float), 3, file);
      if (!toworld)
      {
        fread(&worldpoints[i].pairid, sizeof(int), 1, file);
        if (worldpoints[i].pairid>=0)
          ++matches;
      }
      else
        worldpoints[i].pairid=-1;
    }

    fread(&noofmatchpts, sizeof(int), 1, file);
    if (noofmatchpts>maxmatchpts)
    {
      matchpoints=(POINTDATA*) realloc(matchpoints, noofmatchpts*sizeof(POINTDATA));
      maxmatchpts=noofmatchpts;
    }
    for (i=0; i<noofmatchpts; i++)
    {
      fread(matchpoints[i].xyz, sizeof(float), 3, file);
      fread(matchpoints[i].nxyz, sizeof(float), 3, file);

      fread(&matchpoints[i].pairid, sizeof(int), 1, file);

      if (toworld)
      {
        if (matchpoints[i].pairid>=0)
        {
          assert(matchpoints[i].pairid<noofwldpts);
          worldpoints[matchpoints[i].pairid].pairid=i;
          ++matches;
        }
      }
      else
        fread(&matchpoints[i].index, sizeof(int), 1, file);
    }

    if (!toworld)
    {
      for (i=0; i<noofwldpts; i++)
      {
        if (worldpoints[i].pairid>=0)
        {
          DWORD lb=0, ub=noofmatchpts, mean;

          bool found=false;
          while (ub-lb>0 && !found)
          {
            mean=(ub+lb)/2;
            int comp=worldpoints[i].pairid-matchpoints[mean].index;//_stricmp(word, words[mean]);
            if (comp<0)
              ub=mean;
            else if (comp>0)
              lb=mean+1;
            else
            {
              worldpoints[i].pairid=mean;
              found=true;
            }
          }
          assert(found);
        }
      }
    }

    renderer.SetToWorld(toworld);
    renderer.SetPointData(worldpoints, noofwldpts, false);
    renderer.SetMatchData(matchpoints, noofmatchpts, true);

    wchar_t msg[MAX_PATH];
    wcscpy_s(msg, MAX_PATH, L"Frame No ");
    int len=(int) wcslen(msg);
    _itow_s(frame, msg+len, MAX_PATH-len, 10);
    wcscat_s(msg, MAX_PATH, L": ");
    
    len+=(int) wcslen(msg+len);
    _itow_s(noofwldpts, msg+len, MAX_PATH-len, 10);
    wcscat_s(msg, MAX_PATH, L" world points; ");

    len+=(int) wcslen(msg+len);
    _itow_s(matches, msg+len, MAX_PATH-len, 10);
    wcscat_s(msg, MAX_PATH, L" of ");

    len+=(int) wcslen(msg+len);
    _itow_s(noofmatchpts, msg+len, MAX_PATH-len, 10);
    wcscat_s(msg, MAX_PATH, L" match points (");

    len+=(int) wcslen(msg+len);
    FormatNumber(msg+len, MAX_PATH-len, (100.0*matches)/noofmatchpts, 2);
    wcscat_s(msg, MAX_PATH, L"%) ");

    wcscat_s(msg, MAX_PATH, toworld?L" - to world":L" to frame");
    wcscat_s(msg, MAX_PATH, L"\r\n");

    AppendInfoText(msg);
  }
}

int MainWnd::FormatNumber(wchar_t* buffer, const int bufflen, double value, int dps)
{
  int pos=0;
  if (value<0)
  {
    buffer[0]='-';
    value=-value;
    pos=1;
  }

  int tval=(int) value;
  _itow_s(tval, buffer+pos, bufflen-pos, 10);
  pos+=(int) wcslen(buffer+pos);

  if (dps>0)
  {
    buffer[pos++]='.';
    while (dps>0)
    {
      value=(value-tval)*10;
      tval=(int) value;
      _itow_s(tval, buffer+pos, bufflen-pos, 10);
      pos+=(int) wcslen(buffer+pos);
      --dps;
    }
  }
  buffer[pos]='\0';
  return pos;
}

void MainWnd::CaptureVideo()
{
  wchar_t path[MAX_PATH];
  MultiByteToWideChar(CP_ACP, 0, filename, -1, path, MAX_PATH);

  wchar_t* tmp=wcsrchr(path, '\\');
  if (tmp)
    tmp[1]='\0';
  wcscat_s(path, MAX_PATH, CAPTURE_FOLDER);
  int pathlen=(int) wcslen(path);

  tmp=wcsrchr(path, '\\');
  tmp[0]='\0';
  CreateDirectory(path, NULL);
  tmp[0]='\\';
  
  RECT rect;
  HDC hdcsrc=::GetDC(NULL);
  ::GetWindowRect(renderer.GetHwnd(), &rect);

  int width=rect.right-rect.left;
  int height=rect.bottom-rect.top;

  HDC hdctmp=::CreateCompatibleDC(hdcsrc);

  BITMAPINFO bminfo;
  memset(&bminfo, 0, sizeof(BITMAPINFO));
  bminfo.bmiHeader.biSize=sizeof(BITMAPINFO);
  bminfo.bmiHeader.biWidth=width;
  bminfo.bmiHeader.biHeight=height;
  bminfo.bmiHeader.biPlanes=1;
  bminfo.bmiHeader.biBitCount=32;
  bminfo.bmiHeader.biCompression=BI_RGB;
  bminfo.bmiHeader.biSizeImage=bminfo.bmiHeader.biWidth*bminfo.bmiHeader.biHeight*bminfo.bmiHeader.biBitCount/8;
  
  DWORD* bmpdata=NULL;
  HBITMAP hbmp=::CreateDIBSection(hdcsrc, &bminfo, DIB_RGB_COLORS, (void**) &bmpdata, NULL, NULL);
  ::SelectObject(hdctmp, hbmp);

  BITMAPINFO bmpinfo;
  memset(&bmpinfo, 0, sizeof(BITMAPINFO));
  bmpinfo.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
  bmpinfo.bmiHeader.biWidth=width;
  bmpinfo.bmiHeader.biHeight=height;
  bmpinfo.bmiHeader.biPlanes=1;
  bmpinfo.bmiHeader.biBitCount=24;
  bmpinfo.bmiHeader.biCompression=BI_RGB;
  bmpinfo.bmiHeader.biSizeImage=bmpinfo.bmiHeader.biWidth*bmpinfo.bmiHeader.biHeight*bmpinfo.bmiHeader.biBitCount/8;
  bmpinfo.bmiHeader.biXPelsPerMeter=0;
  bmpinfo.bmiHeader.biYPelsPerMeter=0;
  bmpinfo.bmiHeader.biClrUsed=0;
  bmpinfo.bmiHeader.biClrImportant=0;

  BITMAPFILEHEADER fheader;
  fheader.bfType=0x4D42;
  fheader.bfReserved1=fheader.bfReserved2=0;
  fheader.bfOffBits=sizeof(BITMAPINFO)+sizeof(BITMAPFILEHEADER);
  fheader.bfSize=sizeof(BITMAPINFO)+sizeof(BITMAPFILEHEADER)+bmpinfo.bmiHeader.biSizeImage;

  for (int i=1; i<=noofframes; i++)
  {
    LoadFrame(i);

    int frame=i;
    do
    {
      int sublen=pathlen;
      if (frame<1000)
        path[sublen++]='0';
      if (frame<100)
        path[sublen++]='0';
      if (frame<10)
        path[sublen++]='0';
      _itow_s(frame, path+sublen, MAX_PATH-sublen, 10);
      wcscat_s(path, MAX_PATH, L".bmp");
    
      ++frame;
    } while (GetFileAttributes(path)!=INVALID_FILE_ATTRIBUTES);

    ::BitBlt(hdctmp, 0, 0, width, height, hdcsrc, rect.left, rect.top, SRCCOPY);

    FILE* file;
    _wfopen_s(&file, path, L"wb");
    fwrite(&fheader, sizeof(BITMAPFILEHEADER), 1, file);
    fwrite(&bmpinfo, sizeof(BITMAPINFO), 1, file);
    for (int j=0; j<height; j++)
    {
      int pos=j*width;
      for (int i=0; i<width; i++)
      {
        BYTE* data=(BYTE*) (bmpdata+pos);
        fwrite(data, 1, 1, file);
        fwrite(data+1, 1, 1, file);
        fwrite(data+2, 1, 1, file);
        ++pos;
      }
    }
    fclose(file);

  }

  ::ReleaseDC(NULL, hdcsrc);
  ::DeleteObject(hbmp);
  ::DeleteDC(hdctmp);

}